import torch
import torch.nn as nn
import torch.nn.functional as F

class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)

class Conv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.conv(x)

class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)



class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        self.up = F.interpolate
        self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)


    def forward(self, x1, x2):
        x1 = self.up(x1, x2.shape[2:], mode='bilinear')
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)

class UpInit(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, temp_channels, in_channels, out_channels, bilinear=True, hax=False):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        self.up = F.interpolate
        self.conv1 = Conv(in_channels, temp_channels)
        if hax:
            self.conv2 = DoubleConv(in_channels+temp_channels, out_channels, in_channels // 2)
        else:
            self.conv2 = DoubleConv(in_channels, out_channels, in_channels // 2)


    def forward(self, x1, x2):
        x1 = self.conv1(x1)
        x1 = self.up(x1, x2.shape[2:], mode='bilinear')
        x = torch.cat([x2, x1], dim=1)
        return self.conv2(x)

class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)




import torch
import pdb,os
import torch.nn as nn
import torchvision
import torchvision.models as m
from collections import OrderedDict
BatchNorm = nn.BatchNorm2d
import torch.nn.functional as F

class backbone(nn.Module):
    def __init__(self, network_name):
        super(backbone, self).__init__()
        network_name = network_name.split('_')[1].lower()
        self.network_name = network_name
        self.backbone = self.get_backbone(network_name)        
        self.avgpool = torch.nn.AdaptiveAvgPool2d((1,1))

    def get_backbone(self,network_name):
        if network_name == 'vgg16':
            full_net = getattr(m, network_name)()
            features = list(full_net.features)[:30]
            
            net =  nn.Sequential(*features)

            self.classifier = nn.Sequential(*list(full_net.classifier)[:5])

        elif 'resnet' in network_name:
            full_net = getattr(m, network_name)()

            features = [
                            ('layer0',torch.nn.Sequential(*[full_net.conv1,
                            full_net.bn1,
                            full_net.relu,
                            full_net.maxpool])),
                            ('layer1',full_net.layer1),
                            ('layer2',full_net.layer2),
                            ('layer3',full_net.layer3),
                            ('layer4',full_net.layer4),
                        ]

            self.inplanes =  full_net.inplanes

            net =  nn.Sequential(OrderedDict(features))
        else:
            raise '{} not implemented as BACKBONE Network'.format(network_name)

        return net

    
    def _make_layer(self, block, inplanes, planes, blocks, stride=1, dilation=1, new_level=True, residual=True):
        assert dilation == 1 or dilation % 2 == 0
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm(planes * block.expansion),
            )

        layers = list()
        layers.append(block(
            inplanes, planes, stride, downsample,
            dilation=(1, 1) if dilation == 1 else (
                dilation // 2 if new_level else dilation, dilation),
            residual=residual))
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes, residual=residual,
                                dilation=(dilation, dilation)))

        return nn.Sequential(*layers)

    def insert_dilations(self, features, dilation_size):
        feat = [f for f in features.children()]
        for f in feat:
            if not hasattr(f,'downsample'):
                for g in f.children():
                    tmp = g
                    if type(tmp) == nn.Conv2d:
                        if tmp.kernel_size != (1,1) :
                            tmp.stride   = 1
                            tmp.dilation = (int(dilation_size),int(dilation_size))
                            tmp.padding  = (int(dilation_size),int(dilation_size))
            elif f.downsample is None:
                for g in f.children():
                    tmp = g
                    if type(tmp) == nn.Conv2d:
                        if tmp.kernel_size != (1,1) :
                            tmp.stride   = 1
                            tmp.dilation = (int(dilation_size),int(dilation_size))
                            tmp.padding  = (int(dilation_size),int(dilation_size))
            else:
                if type(f) == torchvision.models.resnet.BasicBlock:
                    for g in f.children():
                        tmp = g
                        if type(tmp) == nn.Conv2d:
                            if tmp.kernel_size != (1,1) :
                                tmp.stride   = 1
                                tmp.dilation = (int(dilation_size),int(dilation_size))
                                tmp.padding  = (int(dilation_size),int(dilation_size))
                        elif type(tmp) == nn.Sequential:
                            for k in tmp.children():
                                if type(k) == nn.Conv2d:
                                    k.stride= 1
                else:
                    for g in f.children():
                        tmp = g
                        if type(tmp) == nn.Conv2d:
                            if tmp.kernel_size != (1,1) :
                                tmp.stride   = 1
                                # tmp.dilation = (int(dilation_size),int(dilation_size))
                                # tmp.padding  = (int(dilation_size),int(dilation_size))
                        elif type(tmp) == nn.Sequential:
                            for k in tmp.children():
                                if type(k) == nn.Conv2d:
                                    k.stride= 1
                                    # k.dilation = (1,1) # (int(dilation_size),int(dilation_size))
                                    # k.padding  = (0,0) # (int(dilation_size),int(dilation_size))
        return features

    def forward(self,x):
        if 'resnet' in self.network_name:
            return self.preset_forward(x)
        else:
            return self._forward(x)

    def _forward(self, x ):
        return self.backbone(x)

    def preset_forward(self, x,  insert_layer = None, return_layer = [1,2,3,4,5]):
        assert (insert_layer is None or return_layer is None or type(return_layer) is list or insert_layer < return_layer )

        if type(return_layer) is int:
            return_layer = [return_layer]
        result = OrderedDict()
        if insert_layer is None or insert_layer == 0:
            x = self.backbone[0](x)
        if  1 in return_layer:
            result['feats_{}_map'.format(1)] = x
            if 1 == max(return_layer):
                return result
        if insert_layer is None or insert_layer <= 1:
            x = self.backbone[1](x)
        if  2 in return_layer:
            result['feats_{}_map'.format(2)] = x
            if 2 == max(return_layer):
                return result
        if insert_layer is None or insert_layer <= 2:
            x = self.backbone[2](x)
        if  3 in return_layer:
            result['feats_{}_map'.format(3)] = x
            if 3 == max(return_layer):
                return result
        if insert_layer is None or insert_layer <= 3:
            x = self.backbone[3](x)
        if  4 in return_layer:
            result['feats_{}_map'.format(4)] = x
            if 4 == max(return_layer):
                return result
        if insert_layer is None or insert_layer <= 4:
            x = self.backbone[4](x)

        result['feats_last_map'] = x

        x = self.avgpool(x)

        x = torch.flatten(x, 1)
        result['feats_pooled']  = x

        return result



# https://github.com/chrise96/image-to-coco-json-converter/tree/master

from PIL import Image                                      # (pip install Pillow)
import numpy as np                                         # (pip install numpy)
from skimage import measure                                # (pip install scikit-image)
import os
import json


def create_category_annotation(category_dict):
    category_list = []

    for key, value in category_dict.items():
        category = {
            "supercategory": key,
            "id": value,
            "name": key
        }
        category_list.append(category)

    return category_list

def get_coco_json_format():
    # Standard COCO format 
    coco_format = {
        "info": {},
        "licenses": [],
        "images": [{}],
        "categories": [{}],
        "annotations": [{}]
    }

    return coco_format


import SimpleITK as sitk
import pydicom
import os
import pydicom_seg
from SimpleITK import ImageSeriesReader
from pydicom import Dataset
from pydicom_seg import MultiClassWriter
import numpy as np


def write_dicom_seg(
                    metainfo: str,
                    # actually a directory, not a full path, though you can adjust it to make it
                    # work with paths
                    dcm_file: str,
                    mask: np.array,
                    out_dir: str,
                    id_label_dict: dict,
                ) -> Dataset:
    """
    :param metainfo: path to metainfo.json
    :param dcm_file: directory to the stored series (or singular file)
    :param mask: segmentation mask
    :return: pydicom dataset containing the dicom seg
    """
    template: Dataset = pydicom_seg.template.from_dcmqi_metainfo(metainfo)
    writer: MultiClassWriter = pydicom_seg.MultiClassWriter(
        template=template,
        inplane_cropping=False,
        skip_empty_slices=True,
        skip_missing_segment=False
    )
    reader: ImageSeriesReader = sitk.ImageSeriesReader()
    dcm_files: tuple = (dcm_file,)
    reader.SetFileNames(dcm_files)
    image = reader.Execute()
    image_data_shape: tuple = sitk.GetArrayFromImage(image).shape

    assert image_data_shape[1:] == mask.shape[1:], f'Mask dimensions {mask.shape} and raw image dimensions {image_data_shape} ' \
                                           f'do not match'
    
    for i in range(mask.shape[0]):
        if mask[i:i+1].sum()==0:
            continue
        segmentation_data = mask[i:i+1]
        segmentation = sitk.GetImageFromArray(segmentation_data)
        segmentation.CopyInformation(image)

        source_images = []
        for file in dcm_files:
            ds = pydicom.dcmread(file, stop_before_pixels=True)
            try:
                _ = ds[0x0020, 0x0032]
            except KeyError:
                ds.ImagePositionPatient = [-212.025, -173.925, -0]
                # raise AttributeError(f'ImagePositionPatient missing in {file}')


            try:
                _ = ds[0x0020, 0x0037]
            except KeyError:
                ds.ImageOrientationPatient = [1, -0, 0, -0, 1, -0]
                # raise AttributeError(f'ImageOrientationPatient missing in {file}')

            source_images.append(ds)

        out = writer.write(segmentation, source_images)
        out_path = os.path.join(out_dir, 
                                id_label_dict[str(i)]\
                               +'.dcm'
                               )
        out.save_as(out_path)



from .cardiothoracic_ratio import get_cardiothoracic_ratio
from .spinecenter_distance import get_spine_center_distance
import numpy as np

class Extractor():
    def __init__(self,):
        """
        """
        self.methods = {
            'Cardio-Thoracic Ratio': get_cardiothoracic_ratio,
            'CTR': get_cardiothoracic_ratio,
            'Spine-Center Distance': get_spine_center_distance,
            'SCD': get_spine_center_distance,
        }
    
    def extract(self, 
                file:np.array, 
                method:str, 
                image: np.array = None,
                draw:bool=False
               ) -> dict:
        """
        """
        assert method in list(self.methods.keys()), f'Method in question ({method}) is not yet implemented. Please write an issue if you want to have it implemented.'
        return self.methods[method](file, image, draw)


import numpy as np
from PIL import Image
from cxas.label_mapper import label_mapper
from cxas.extraction.draw_helpers import draw_point, draw_line

def get_cardiothoracic_ratio(npy, img=None, draw = False):
    """
    Calculate Cardio-thoracic-ratio from mask (check https://en.wikipedia.org/wiki/Cardiomegaly)
    MRD = greatest perpendicular diameter from midline to right heart border
    MLD = greatest perpendicular diameter from midline to left heart border
    ID = internal diameter of chest at level of right hemidiaphragm

    Parameters
    ----------
        npy: mask in form of np array [n_classes, width, height]
        img: source image, only used for visualization
        draw: whether to visualize the features
        
    Returns
    -------
        CTR: (MRD + MLD) / ID
    """

    def get_longest(mask):
        min_ = 0
        id_ = 0
        points = (0,0),(0,0)
        for i in range(len(mask)):
            if mask[i].sum() == 0:
                continue
            nz = mask[i].nonzero()[0]
            if len(nz)!=0:
                dist = nz.max() - nz.min()
                if dist>min_:
                    min_ = dist
                    id_ = i
                    points = [(i,nz.min()), (i,nz.max())]
        return min_, id_, points
    
    def distance_to_midline(midline, heart):
        mrd = 0
        mld = 0
        mrd_id = 0
        mld_id = 0
        min_pos, max_pos = (0,0),(0,0)
        midline1, midline2 = (0,0), (0,0)
        for i in range(len(heart)):
            if heart[i].sum() == 0:
                continue
            nz = heart[i].nonzero()[0]
            if len(nz)!=0:
                dist_ld = nz.max() - midline
                dist_rd = midline - nz.min()

                if dist_rd > mrd:
                    mrd = dist_rd
                    mrd_id = i
                    min_pos = (i,nz.min())
                    midline1 = (i, midline)
                if dist_ld > mld:
                    mld = dist_ld
                    mld_id = i
                    max_pos = (i,nz.max())
                    midline2 = (i, midline)
        return mrd, mld, mrd_id, mld_id, min_pos, max_pos, midline1, midline2
            
    if (npy[label_mapper['lung'][0]].sum()==0) or\
        (npy[label_mapper['right hemidiaphragm'][0]].sum()==0) or\
        (npy[label_mapper['heart'][0]].sum()==0):
        return {'score': -1, 'drawing': Image.new("RGB", (npy.shape[1], npy.shape[2]), "black")}
    
    lung = npy[label_mapper['lung'][0]]
    midline = npy[label_mapper['right hemidiaphragm'][0]].nonzero()[1].max()
    heart = npy[label_mapper['heart'][0]]
    
    mrd, mld, _, _, min_pos, max_pos, midline1, midline2 = distance_to_midline(midline, heart)
    ID, _, points = get_longest(lung)
    
    
    
    
    if draw:
        if img is None:
            img = Image.new("RGB", (npy.shape[1], npy.shape[2]), "black")
        else:
            pass
        
        width = 8

        img = draw_line(img, (points[0][1],points[0][0]), (points[1][1],points[1][0]), '#2A9D8F',width)
        img = draw_point(img, (points[0][1],points[0][0]), '#264653', width*2)
        img = draw_point(img, (points[1][1],points[1][0]), '#264653', width*2)


        img = draw_line(img, (min_pos[1],min_pos[0]), (midline1[1],midline1[0]), '#F4A261',width)
        img = draw_line(img, (midline2[1],midline2[0]), (midline1[1],midline1[0]), '#F4A261',width)
        img = draw_line(img, (max_pos[1],max_pos[0]), (midline2[1],midline2[0]), '#F4A261',width)

        img = draw_point(img, (min_pos[1],min_pos[0]), '#E76F51', width*2)
        img = draw_point(img, (max_pos[1],max_pos[0]), '#E76F51', width*2)
        img = draw_point(img, (midline1[1],midline1[0]), '#E76F51', width*2)
        img = draw_point(img, (midline2[1],midline2[0]), '#E76F51', width*2)
        
        return {'score':(mrd+mld)/ID, 'drawing':img}
    return {'score':(mrd+mld)/ID}


import numpy as np
import json
from pycocotools import mask

# Function to convert binary mask to COCO RLE format
def binary_mask_to_rle(
                        binary_mask:np.array,
                    ):
    """
    """
    mask_encoded = mask.encode(np.asfortranarray(binary_mask.astype(np.uint8)))
    mask_encoded['counts'] = mask_encoded['counts'].decode('utf-8')
    return mask_encoded

def mask_to_annotation(
                        mask:np.array, 
                        base_ann_id:int=1,
                        img_id:int = 1,
                        ):
    """
    """
    annotations = []
    for i in range(mask.shape[0]):
        if mask[i].sum()==0:
            continue
        binary_mask = mask[i]
        # import pdb;pdb.set_trace()
        mask_encoded = binary_mask_to_rle(binary_mask)
        annotation = {
            'id': base_ann_id + i,
            'image_id': img_id,  # Assuming all masks belong to the same image
            'category_id': i,  # Assuming a single category
            'segmentation': mask_encoded,
            'area': int(np.sum(binary_mask)),
            'bbox': toBox(mask_encoded).tolist(),
            'iscrowd': 0  # Set to 1 if the mask represents a crowd region
        }
        annotations.append(annotation)
    return annotations


def toBox(binary_mask):
    return mask.toBbox(binary_mask)


from PIL import Image
import SimpleITK as sitk 
import numpy as np
import os
import json 
import torch
import torch.nn.functional as F
from itertools import groupby
from cxas.label_mapper import id2label_dict, category_ids
from torch.utils.data import Dataset
import SimpleITK as sitk
import numpy as np
from cxas.io_utils.dicomseg_2d import write_dicom_seg
from cxas.io_utils.create_annotations import get_coco_json_format, \
    create_category_annotation
from cxas.io_utils.mask_to_coco import toBox, mask_to_annotation
from pathlib import Path

this_directory = Path(__file__).parent


class FolderDataset(Dataset):
    def __init__(self, path:str, gpus:str):
        super(Dataset,self).__init__()
        file_types = ['jpg', 'png', 'dcm']
        self.fileloader = FileLoader('')
        self.files = [os.path.join(path, i) for i in os.listdir(path) if i.split('.')[-1].lower() in file_types]
    
    def collate_fn(self, batch):
        out_dict = {}
        for b in batch:
            for k in b.keys():
                if k in out_dict.keys():
                    out_dict[k] += [b[k]]
                else:
                    out_dict[k] = [b[k]]


        for k in out_dict.keys():
            if k=='data':
                out_dict[k] = torch.cat(out_dict[k],dim=0)
                       
        return out_dict
        
    def __getitem__(self, index):
        return self.fileloader.load_file(self.files[index])
    
    def __len__(self,):
        return len(self.files)

def get_folder_loader(path:str, gpus:str, batch_size:int)-> torch.utils.data.DataLoader:
    dataset = FolderDataset(path, gpus)
    loader = torch.utils.data.DataLoader(dataset,
                      batch_size=batch_size,
                      shuffle=False,
                      num_workers=4,
                      collate_fn= dataset.collate_fn)
    return loader
    
class FileLoader():
    def __init__(self, 
                 gpus:str
                ):
        """
        """
        self.base_size = 512
        self.file_types = {
            'jpg': self.load_image,
            'png': self.load_image,
            'dcm': self.load_dicom,
        }

        self.gpus = [int(i) for i in gpus.split(',') if len(i)>0]
        
    def to_gpu(self, 
               array: torch.tensor
              )-> torch.tensor:
        """
        :param array:
        :return: 
        """
        if len(self.gpus)>0:
            assert (torch.cuda.is_available())
            return array.to(torch.device('cuda:{}'.format(self.gpus[0])))
        else:
            return array
        
    def normalize(self, 
                  array:torch.tensor
                 )-> torch.tensor:
        """
        """
        assert array.shape[0] == 3, f'{array.shape}'
        # ImageNet normalization
        # Array to be assumed in range [0,1]
        assert (array.min()>=0) and (array.max()<=1)
        
        array = (array - torch.tensor([0.485, 0.456, 0.406]).view(-1,1,1))/(torch.tensor([0.229, 0.224, 0.225]).view(-1,1,1))
        return array
    
    def load_file(self, 
                  file_path: str
                 )-> dict:
        
        assert file_path.split('.')[-1].lower() in list(self.file_types.keys()), f"filetype not supported: {file_path.split('.')[-1].lower()}"
        
        return self.file_types[file_path.split('.')[-1].lower()](file_path)
        
    def load_image(self, 
                   image_path: str
                  ) -> dict:
        """
        """
        array = np.array(Image.open(image_path).convert(mode='RGB'))
        array = np.transpose(array, [2,0,1])
        original_array = np.copy(array)
        orig_file_size = array.shape[-2:]
        array = torch.tensor(array).float()/255
        array = self.normalize(array)
        array = self.to_gpu(array)
        return {
                'data':F.interpolate(array.unsqueeze(0), self.base_size),
                'orig_data': original_array,
                'filename': image_path,
                'file_size': orig_file_size
               }

    def load_dicom(self, 
                   image_path: str
                  ) -> dict:
        """
        """
        image = sitk.ReadImage(image_path)
        array_view = sitk.GetArrayFromImage(image).astype(np.float32)
        assert (len(array_view.shape) == 3) and (array_view.shape[0] ==1)
        orig_file_size = array_view.shape[-2:]
        tensor = torch.tensor(array_view).float()
        tensor = (tensor-tensor.min())/(tensor.max()-tensor.min())
        tensor = torch.cat([tensor,tensor,tensor],0)
        original_array = (tensor.numpy()*255).astype(np.uint8)
        array = self.normalize(tensor)
        array = self.to_gpu(array)
        return {
                'data':F.interpolate(array.unsqueeze(0), self.base_size),
                'orig_data': original_array,
                'filename': image_path,
                'file_size': orig_file_size
               }
        
        
    
class FileSaver():   
    def __init__(self,):
        """
        """
        self.save_modes = {
            'npz': self.export_prediction_as_npz,
            'npy': self.export_prediction_as_numpy,
            'jpg': self.export_prediction_as_jpg,
            'png': self.export_prediction_as_png,
            'json': self.export_prediction_as_json,
            'dicom-seg': self.export_prediction_as_dicomseg,
            }
        
    def save_prediction(self, 
                        mask:np.array, 
                        outdir:str,  
                        file_name:str, 
                        mode: str
                       )-> None:
        assert mode in list(self.save_modes.keys())
        assert len(mask.shape) ==3
        self.save_modes[mode](mask, outdir, file_name)      
    
    def export_prediction_as_dicomseg(self, 
                                      mask:np.array, 
                                      outdir:str,  
                                      file_name:str
                                     ) -> None:
        """
        """
        assert os.path.splitext(file_name)[-1] == 'dcm', 'Input file has to be dicom to be stored \
            as dicom-seg, it was --{}--'.format(os.path.splitext(file_name)[-1])
        out_dir = os.path.join(outdir, os.path.splitext(file_name)[0])
        os.makedirs(out_dir,exist_ok=True)
        
        write_dicom_seg(
                            metainfo = os.path.join(str(this_directory), 'data/metainfo.json'),
                            dcm_file = file_name,
                            mask = mask.astype(np.uint8),
                            out_dir = out_dir,
                            id_label_dict = id2label_dict,
                            )
                
    
    def export_prediction_as_jpg(self, 
                                 mask:np.array, 
                                 outdir:str,  
                                 file_name:str
                                ) -> None:
        """
        """
        assert len(mask.shape) == 3
        fileroot = os.path.splitext(file_name)[0]
        outdir = os.path.join(outdir, fileroot)
        os.makedirs(outdir,exist_ok=True)
        
        for i in range(len(mask)):
            out_path = os.path.join(outdir, id2label_dict[str(i)] + '.jpg')
            Image.fromarray(mask[i]).convert('1').save(out_path)
    
    def export_prediction_as_png(self, 
                                 mask:np.array, 
                                 outdir:str,  
                                 file_name:str
                                ) -> None:
        """
        """
        assert len(mask.shape) == 3
        fileroot = os.path.splitext(file_name)[0]
        outdir = os.path.join(outdir, fileroot)
        os.makedirs(outdir,exist_ok=True)
        
        for i in range(len(mask)):
            out_path = os.path.join(outdir, id2label_dict[str(i)] + '.png')
            Image.fromarray(mask[i]).convert('1').save(out_path)

    def export_prediction_as_numpy(self, 
                                   mask:np.array, 
                                   outdir:str,  
                                   file_name:str
                                  ) -> None:
        """
        """
        os.makedirs(outdir,exist_ok=True)
        out_path = os.path.join(outdir, os.path.splitext(file_name)[0]+'.npy')
        np.save(out_path, mask)
    
    def export_prediction_as_npz(self, 
                                 mask:np.array,
                                 outdir:str,  
                                 file_name:str
                                ) -> None:
        os.makedirs(outdir,exist_ok=True)
        out_path = os.path.join(outdir, os.path.splitext(file_name)[0]+'.npz')
        np.savez_compressed(out_path, mask)

    
    def export_prediction_as_json(self, 
                                  mask:np.array, 
                                  outdir:str,  
                                  file_name:str,
                                  img_id:int=1,
                                  base_ann_id:int=1,
                                 ) -> None:
        
        coco_format = get_coco_json_format()
        coco_format["categories"] = create_category_annotation(category_ids)
        
        coco_format["images"] = []
        coco_format["images"].append([{'id':img_id, 'file_name': file_name}])
        
        coco_format["annotations"] = mask_to_annotation(
                                        mask = mask, 
                                        base_ann_id = base_ann_id, 
                                        img_id = img_id
                                    ) 
        
        os.makedirs(outdir,exist_ok=True)
        out_path = os.path.join(outdir, os.path.splitext(file_name)[0]+'.json')
        
        with open(out_path,"w") as outfile:
            json.dump(coco_format, outfile)


import torch
import torch.nn as nn
import torch.nn.functional as F
from .backbones import backbone
from .unet_components import *

class BackboneUNet(nn.Module):
    def __init__(self, model_name:str, classes:int):
        super(BackboneUNet,self).__init__()
        
        self.backbone  = backbone(model_name)
        self.head      = get_unet_head(model_name, classes)
        self.dropout = nn.Dropout(p=0)
        self.threshold = 0.5

    def get_results(self, forward_dict, orig_dict):
        out_dict = {**forward_dict,**orig_dict}

        out_dict['segmentation_preds'] = (forward_dict['logits'].sigmoid()>self.threshold).bool()
        
        return out_dict

    def forward(self,x):
        img = x['data']
        forward_dict = self._forward(img)
        out_dict = self.get_results(forward_dict, x)
        return out_dict

    def _forward(self, x ):
        backbone_dict = self.backbone(x)

        down1 = backbone_dict['feats_1_map']
        down2 = backbone_dict['feats_2_map']
        down3 = backbone_dict['feats_3_map']
        down4 = backbone_dict['feats_4_map']
        down5 = backbone_dict['feats_last_map']
        # import pdb;pdb.set_trace()
        up1    = self.dropout(self.head.up1(down5, down4))
        up2    = self.dropout(self.head.up2(up1, down3))
        up3    = self.dropout(self.head.up3(up2, down2))
        up4    = self.dropout(self.head.up4(up3, down1))
        
        up     = F.interpolate(up4, x.shape[2:], mode='bilinear')
        logits = self.head.out(up)
        
        return {'feats':up4,'logits': logits}

def get_unet_head(network_name, classes, batch_size=1):
    if 'vit' in network_name:
        return UNetHead(256, 128, classes, norm = 'batch' if batch_size >1 else 'instance', constant= True)
    elif 'resnet34' in network_name:
        return UNetHead(512, 128, classes, norm = 'batch' if batch_size >1 else 'instance')
    else:
        return UNetHead(2048, 128, classes, norm = 'batch' if batch_size >1 else 'instance')

class UNetHead(nn.Module):
    def __init__(self, in_channels, ngf, num_classes, norm='batch', constant = False):
        super(UNetHead, self).__init__()

        if constant:
            self.up1 = UpInit(in_channels,in_channels,in_channels, True, True)
            self.up2 = Up(2*in_channels,in_channels, True)
            self.up3 = Up(2*in_channels,in_channels, True)
            self.up4 = Up(2*in_channels,in_channels, True)

            self.out = OutConv(in_channels, num_classes)

        else:
            self.up1 = UpInit(in_channels //2, in_channels, in_channels //4, True)
            self.up2 = Up(in_channels // 4 * 2,  in_channels //8, True)
            self.up3 = Up(in_channels // 8 * 2,  in_channels // 32, True)
            self.up4 = Up(in_channels // 32 * 2,  ngf, True)

            self.out = OutConv(ngf, num_classes)


import gdown, os, torch
from ..label_mapper import id2label_dict

model_urls = {
    'UNet_ResNet50_default': 'https://drive.google.com/file/d/1Y9zubvMzkYHoAqz-NvV6vniH5FKAF2iV/view?usp=drive_link'
}

def get_model(model_name, gpus=''):
    assert model_name.split('_')[0] in list(model_getter.keys())
    
    model = model_getter[model_name.split('_')[0]](model_name)
    download_weights(model_name)
    model = load_weights(model, model_name, map_location = 'cpu')
    
    if 'cpu' in gpus:
        model.to(torch.device('cpu'))
    else:
        gpus = [int(i) for i in gpus.split(',') if len(i)>0]

        if len(gpus) > 1:
            assert (torch.cuda.is_available())
            model.to(gpus[0])
            model = torch.nn.DataParallel(model,device_ids =gpus).cuda()
        elif len(gpus)==1:
            assert (torch.cuda.is_available())
            model.to(gpus[0])
        
    return model


def get_unet(model_name):
    from .UNet.backbone_unet import BackboneUNet
    return BackboneUNet(model_name, len(id2label_dict.keys()))

def download_weights(model_name:str)->None:
    if "CXAS_PATH" in os.environ:
        store_path = os.path.join(os.environ['CXAS_PATH'],'.cxas')
    else:
        store_path = os.path.join(os.environ['HOME'],'.cxas')
    os.makedirs(os.path.join(store_path, 'weights/'), exist_ok=True)
    out_path = os.path.join(store_path, 'weights/{}'.format(model_name+'.pth'))
    if os.path.isfile(out_path):
        return
    else:
        gdown.download(model_urls[model_name], out_path, quiet=False, fuzzy=True)
        return

def load_weights(model, model_name:str, map_location:str='cuda:0'):
    if "CXAS_PATH" in os.environ:
        store_path = os.path.join(os.environ['CXAS_PATH'],'.cxas')
    else:
        store_path = os.path.join(os.environ['HOME'],'.cxas')
    out_path = os.path.join(store_path, 'weights/{}'.format(model_name+'.pth'))
    assert os.path.isfile(out_path)
    
    checkpoint = torch.load(out_path, map_location=map_location)
    # import pdb; pdb.set_trace()
    if 'module' in list(checkpoint['model'].keys())[0] :
        for i in list(checkpoint['model'].keys()):
            checkpoint['model'][i[len('module.'):]] = checkpoint['model'].pop(i)
    model.load_state_dict(checkpoint['model'], strict = False)
    return model

model_getter = {
    'UNet': get_unet,
}